@@ -3,19 +3,22 @@ use strict;
 use warnings;
 
 package  App::HTTP_Proxy_IMP;
-use fields qw(addr impns filter pcapdir);
+use fields qw(addr impns filter pcapdir mitm_ca capath no_check_certificate);
 
-use App::HTTP_Proxy_IMP::IMP;
+use App::HTTP_Proxy_IMP::IMP '$IMP_MAX_IN_ANALYZER';
 use App::HTTP_Proxy_IMP::Conn;
 use App::HTTP_Proxy_IMP::Request;
 use App::HTTP_Proxy_IMP::Relay;
 use AnyEvent;
 use Getopt::Long qw(:config posix_default bundling);
-use App::HTTP_Proxy_IMP::Debug qw(debug $DEBUG $DEBUG_RX);
+use App::HTTP_Proxy_IMP::Debug ();
+use Net::IMP::Debug qw(debug $DEBUG $DEBUG_RX);
 use Net::Inspect::Debug qw(%TRACE);
+use IO::Socket::SSL::Intercept;
+use IO::Socket::SSL::Utils;
 use Carp 'croak';
 
-our $VERSION = '0.91_2';
+our $VERSION = '0.945_001';
 
 # try IPv6 using IO::Socket::IP or IO::Socket::INET6
 # fallback to IPv4 only
@@ -35,7 +38,7 @@ sub new {
     my ($class,@args) = @_;
     my $self = fields::new($class);
     $self->{impns} = [qw(App::HTTP_Proxy_IMP::IMP Net::IMP::HTTP Net::IMP)];
-    %$self = %{ shift(@args) } if @args && ref($args[0]);
+    %$self = ( %$self, %{ shift(@args) }) if @args && ref($args[0]);
     $self->getoptions(@args) if @args;
     return $self;
 }
@@ -50,6 +53,63 @@ sub start {
 	eval { require Net::PcapWriter } or croak(
 	    "cannot load Net::PcapWriter, which is needed with --pcapdir option");
     }
+
+    my $mitm;
+    if ( my $f = $self->{mitm_ca} ) {
+	my $serial = 1;
+	my $cache = {};
+	my $cachedir = "$f.cache";
+	if ( -d $cachedir || mkdir($cachedir,0700)) {
+	    for my $f (glob("$cachedir/*.pem")) {
+		-f $f && -r _ && -s _ or next;
+		my $time = (stat(_))[9];
+		my $key  = PEM_file2key($f) or next;
+		my $cert = PEM_file2cert($f) or next;
+		my $sn = CERT_asHash($cert)->{serial};
+		$serial = $sn+1 if $sn>=$serial;
+		my ($id) = $f=~m{/([^/]+)\.pem$};
+		$cache->{$id} = {
+		    cert => $cert,
+		    key => $key,
+		    atime => $time,
+		};
+		debug("loaded certificate id=$id from cache");
+	    }
+
+	    my $cache_hash = $cache;
+	    $cache = sub {
+		my $id = shift;
+		my $e;
+		if ( ! @_ ){ # get
+		    $e = $cache_hash->{$id} or return;
+		} else {
+		    my ($cert,$key) = @_;
+		    $e = $cache_hash->{$id} = {
+			cert => $cert,
+			key => $key,
+		    };
+		}
+		my $f = "$cachedir/$id.pem";
+		if ( @_ || ! -f $f and open( my $fh,">",$f )) {
+		    debug("save mitm certificate and key to $cachedir/$id.pem");
+		    print $fh PEM_cert2string($e->{cert}),
+			PEM_key2string($e->{key})
+		} else {
+		    utime(undef,undef,$f);
+		}
+		$e->{atime} = time();
+		return ($e->{cert},$e->{key});
+	    };
+	}
+
+	$mitm = IO::Socket::SSL::Intercept->new(
+	    proxy_cert_file => $f,
+	    proxy_key_file  => $f,
+	    cache => $cache,
+	    serial => $serial,
+	);
+    }
+
     my $imp_factory;
     my $filter = $self->{filter};
     if ($filter && @$filter ) {
@@ -62,6 +122,9 @@ sub start {
 		next;
 	    }
 
+	    my $f = $f; # copy
+	    my $args = $f =~s{=(.*)}{} && $1;
+
 	    my $found;
 	    for my $prefix ('', map { "${_}::" } @$ns) {
 		my $mod = $prefix.$f;
@@ -71,41 +134,72 @@ sub start {
 		}
 	    }
 	    croak("IMP module $f could not be loaded: $@") if ! $found;
-	    push @mod,$found;
+	    my %args = $args ? $found->str2cfg($args) :();
+	    my @err = $found->validate_cfg(%args);
+	    die "bad config for $found: @err" if @err;
+	    push @mod, $found->new_factory(%args)
 	}
 	$imp_factory = App::HTTP_Proxy_IMP::IMP->new_factory(@mod);
     }
 
+    my $capath = $self->{capath};
+    if ( $self->{no_check_certificate} ) {
+	$capath = undef;
+    } elsif ( ! $capath ) {
+	if ( eval { require Mozilla::CA } ) {
+	    $capath = Mozilla::CA::SSL_ca_file();
+	} elsif ( glob("/etc/ssl/certs/*.pem") ) {
+	    $capath = "/etc/ssl/certs";
+	} elsif ( -f "/etc/ssl/certs.pem" && -r _ && -s _ ) {
+	    $capath = "/etc/ssl/certs.pem";
+	} else {
+	    croak "cannot determine CA path, needed for SSL interception"
+	}
+    }
+
     # create connection fabric, attach request handling
     my $req  = App::HTTP_Proxy_IMP::Request->new;
     my $conn = App::HTTP_Proxy_IMP::Conn->new($req, 
 	pcapdir     => $pcapdir, 
+	mitm        => $mitm,
+	capath      => $capath,
 	imp_factory => $imp_factory
     );
 
-    # add hooks for unchunking and decompression
-    # FIXME - add these hooks dynamically if we want to inspect content
-    # and not only header
-    $req->add_hooks('unchunk','uncompress_te','uncompress_ce');
-
-
     # create listeners
     my @listen;
-    $self->{addr} = [ $self->{addr} ] if $self->{addr} && ! ref($self->{addr});
-    for my $addr (@{$self->{addr}}) {
-	my $srv = $sockclass->new(
-	    LocalAddr => $addr,
-	    Listen    => 10,
-	    ReuseAddr => 1,
-	) or croak("cannot listen to $addr: $!");
+    $self->{addr} = [ $self->{addr} ] 
+	if $self->{addr} && ref($self->{addr}) ne 'ARRAY';
+    for my $spec (@{$self->{addr}}) {
+	my ($addr,$upstream) = 
+	    ref($spec) eq 'ARRAY' ? @$spec:
+	    ref($spec) ? ( $spec,undef ):
+	    split('=',$spec,2);
+	my $srv;
+	if ( ref($addr)) {
+	    # listing socket already
+	    $srv = $addr;
+	    (my $port,$addr) = AnyEvent::Socket::unpack_sockaddr( getsockname($srv));
+	    $addr = AnyEvent::Socket::format_address($addr);
+	    $addr = $addr =~m{:} ? "[$addr]:$port" : "$addr:$port";
+	} else {
+	    $srv = $sockclass->new(
+		LocalAddr => $addr,
+		Listen    => 10,
+		ReuseAddr => 1,
+	    ) or croak("cannot listen to $addr: $!");
+	}
+	$spec = [ $addr,$upstream,$srv ];
 	push @listen, AnyEvent->io(
 	    fh => $srv,
 	    poll => 'r',
 	    cb => sub {
 		my $cl = $srv->accept or return;
-		App::HTTP_Proxy_IMP::Relay->new($cl,$conn);
+		debug("new request from %s:%s on %s",$cl->peerhost,$cl->peerport,$addr);
+		App::HTTP_Proxy_IMP::Relay->new($cl,$upstream,$conn);
 	    }
 	);
+	debug("listening on $addr");
     }
 
     # on SIGUSR1 dump state of all relays
@@ -142,7 +236,10 @@ sub getoptions {
     local @ARGV = @_;
     GetOptions(
 	'h|help'      => sub { usage() },
-	'P|pcapdir=s' => sub { $self->{pcapdir} = $_[1] },
+	'P|pcapdir=s' => \$self->{pcapdir},
+	'mitm-ca=s'   => \$self->{mitm_ca},
+	'capath=s'    => \$self->{capath},
+	'no-check-certificate=s' => \$self->{no_check_certificate},
 	'F|filter=s'  => sub { 
 	    if ($_[1] eq '-') { 
 		# discard all previously defined
@@ -170,6 +267,7 @@ sub getoptions {
 	'T|trace=s' => sub { 
 	    $TRACE{$_} = 1 for split(m/,/,$_[1]) 
 	},
+	'max-in-analyzer=i' => \$IMP_MAX_IN_ANALYZER,
     );
 
     my @addr = @ARGV;
@@ -188,10 +286,22 @@ sub usage {
 HTTP proxy, which can inspect and modify requests and responses before
 forwarding using Net::IMP plugins.
 
-$cmd Options* [ip:port]+
+$cmd Options* [ip:port|ip:port=upstream_ip:port]+
 ip:port - listen address(es) for the proxy
+ip:port=upstream_ip:port - listen adress and upstream proxy 
+
 Options:
   -h|--help        show usage
+
+  --mitm-ca ca.pem use given file in PEM format as a Proxy-CA for intercepting
+                   SSL connections (e.g. man in the middle). Should include key
+		   and cert.
+  --capath P       path to file or dir containing CAs, which are used to verify
+                   server certificates when intercepting SSL.
+		   Tries to use builtin default if not given.
+  --no-check-certificate  do not check server certificates when intercepting
+                   SSL connections
+
   -F|--filter F    add named IMP plugin as filter, can be used multiple times
                    with --filter mod=args arguments can be given to the filter
   --imp-ns N       perl module namespace, were it will look for IMP plugins.
@@ -199,6 +309,10 @@ Options:
 		   Plugins outside these namespace need to be given with 
 		   full name.
 		   Defaults to App::HTTP_Proxy_IMP, Net::IMP
+
+  --max-in-analyzer L   set limit, on how many bytes may be send to IMP
+                   analyzer w/o getting feedback from analyzer, default 2^20.
+		   Set to 0 for unlimited
   
   # options intended for development and debugging:
   -P|--pcapdir D   save connections as pcap files into D, needs Net::PcapWriter
@@ -213,6 +327,9 @@ start proxy at 127.0.0.1:8888 and log all form fields
  $cmd --filter LogFormData 127.0.0.1:8888
 start proxy at 127.0.0.1:8888 with CSRF protection plugin
  $cmd --filter CSRFprotect 127.0.0.1:8888
+start proxy at 127.0.0.1:8888 with CSRF protection plugin, using upstream 
+proxy proxy:8888
+ $cmd --filter CSRFprotect 127.0.0.1:8888=proxy:8888
 
 USAGE
     exit(2);
@@ -240,7 +357,7 @@ App::HTTP_Proxy_IMP - HTTP proxy with the ability to inspect and modify content
 
     # show cmdline usage
     App::HTTP_Proxy_IMP->usage();
-    
+
 =head1 DESCRIPTION
 
 App::HTTP_Proxy_IMP implements an HTTP proxy, which can inspect and modify the
@@ -288,14 +405,63 @@ The cmdline option can be given multiple times.
 If '-' is given at cmdline all previously defined prefixes (including defaults) are
 discarded.
 
-=item addr ARRAY|ip:port
+=item addr [spec+]
 
-List of ip:port combinations or single ip:port.
-These are the local addresses where the proxy will listen.
+Array of listener/upstream specifications for proxy.  
+Each specification can be
+
+=over 12
+
+=item ip:port - address where the proxy should listen
+
+=item ip:port=target_ip:port - listener address and upstream proxy address
+
+=item socket - precreated listener socket
+
+=item [ socket, target_ip:port ] - precreated listener socket and address of
+  upstream proxy
+
+=back
 
 On the cmdline these are given as the remaining arguments, e.g. after all
 other options.
 
+=item mitm_ca proxy_ca.pem
+
+When this parameter is given it will intercept SSL connections by ending the
+connection from the server at the proxy and creating a new connection with a
+new certificate signed by the given ca.pem (e.g. man in the middle).
+Thus it will be able to analyse and manipulate encrypted connections.
+
+ca.pem needs to include the CA cert and key in PEM format.
+Also you better import the CA certificate into your browser, or you will get
+warnings on access to SSL sites, because of the correctly detected man in the
+middle attack.
+
+To generate the proxy certificate you might use openssl:
+
+   openssl genrsa -out key.pem 1024
+   openssl req -new -x509 -extensions v3_ca -key key.pem -out proxy_ca.pem 
+   cat key.pem >> proxy_ca.pem
+   # export to PKCS12 for import into browser
+   openssl pkcs12 -export -in proxy_ca.pem -inkey proxy_ca.pem -out proxy_ca.p12
+
+It will try to create the directory proxy_ca.pem.cache and use it as a cache
+for generated cloned certificates. If this is not possible the cloned certificates
+will persist over restarts of the proxy.
+
+=item capath certs.pem | cert-dir
+
+The path (file with certificates or directory) with the CAs, which are used to
+verify SSL certificates when doing SSL interception.
+If not given it will check initially for various path, starting with using
+Mozilla::CA, trying /etc/ssl/certs and /etc/ssl/certs.pem before giving up and
+exiting.
+
+=item no_check_certificate
+
+If true disables checking of SSL certificates when doing SSL interception.
+
 =back
 
 The following options are only for the cmdline
@@ -319,7 +485,7 @@ L<Net::Inspect::Debug> package.
 
 =back
 
-=item start
+=item * start
 
 Start the proxy, e.g. start listeners and process incoming connections.
 No arguments are expected if called on an object, but one can use the form
@@ -331,6 +497,8 @@ loop using C<< AnyEvent->condvar->recv >>.
 If a value is expected it will return 1, and the caller hast to enter the
 AnyEvent mainloop itself.
 
+=back
+
 =head2 Reaction to Signals
 
 The installs some signal handlers:
