@@ -5,10 +5,12 @@ package App::HTTP_Proxy_IMP::Relay;
 use fields (
     'fds',      # file descriptors
     'conn',     # App::HTTP_Proxy_IMP::HTTPConn object
+    'acct',     # collect accounting
 );
 
 use App::HTTP_Proxy_IMP::Debug;
 use Scalar::Util 'weaken';
+use IO::Socket::SSL;
 use AnyEvent;
 
 # active relay, inserted in new, removed in $idlet timer
@@ -17,15 +19,22 @@ sub relays { return grep { $_ } @relays }
 
 # creates new relay and puts it into @relays as weak reference
 sub new {
-    my ($class,$cfd,$conn) = @_;
+    my ($class,$cfd,$upstream,$conn) = @_;
     my $self = fields::new($class);
     debug("create relay $self");
 
+    if ( $upstream && ! ref($upstream)) {
+	$upstream =~m{\A(?:\[([a-f\d:.]+)\]|([\da-z_\-.]+)):(\d+)\Z} or
+	    die "invalid upstream specification: $upstream";
+	$upstream = [ $1||$2, $3 ];
+    }
+
     my $cobj = $conn->new_connection({
 	daddr => $cfd->sockhost,
 	dport => $cfd->sockport,
 	saddr => $cfd->peerhost,
 	sport => $cfd->peerport,
+	upstream => $upstream,
     },$self);
 
     #debug("create connection $cobj");
@@ -41,16 +50,43 @@ sub new {
 
 sub DESTROY {
     my $self = shift;
+    $self->account('destroy');
     $self->xdebug("destroy relay $self");
 }
 
+sub acctinfo {
+    my ($self,$acct) = @_;
+    $self->{acct} = $acct;
+}
 sub account {
-    my ($self,%args) = @_;
-    if ( my $t = delete $args{start} ) {
-	$args{duration} = AnyEvent->now - $t;
+    my ($self,$what,%args) = @_;
+    my $acct = $self->{acct};
+    $acct = $acct ? { %$acct,%args } : \%args if %args;
+    $acct or return;
+    $self->{acct} = undef;
+    if ( my $t = delete $acct->{start} ) {
+	$acct->{duration} = AnyEvent->now - $t;
     }
-    my $msg = join(' ', map { "$_=$args{$_}" } sort keys %args);
-    print STDERR "ACCT $msg\n";
+    my @msg;
+    for( sort keys %$acct ) {
+	my $t;
+	my $v = $acct->{$_};
+	if ( ! defined $v ) {
+	    next;
+	} elsif ( ref($v) eq 'ARRAY') {
+	    $t = "$_=[".join(',',map { _quote($_) } @$v)."]";
+	} elsif ( defined $v ) {
+	    $t = "$_="._quote($v);
+	}
+	push @msg,$t;
+    }
+    print STDERR "ACCT @msg\n";
+}
+
+sub _quote {
+    my $text = shift;
+    $text =~s{([\000-\037\\"\377-\777])}{ sprintf("\\%03o",ord($1)) }eg;
+    return $text =~m{ } ? qq["$text"]:$text;
 }
 
 sub xdebug {
@@ -60,14 +96,22 @@ sub xdebug {
 	unshift @_,$conn;
 	goto &$xdebug;
     } else {
-	debug(@_);
+	goto &debug;
     }
 }
 
 
+# non-fatal problem
+sub error {
+    my ($self,$reason) = @_;
+    warn "[error] ".( $self->{conn} && $self->{conn}->id || 'noid')." $reason\n";
+    return 0;
+}
+
+# fatal problem - close connection
 sub fatal {
     my ($self,$reason) = @_;
-    warn "[fatal] ".$self->{conn}->id." $reason\n";
+    warn "[fatal] ".( $self->{conn} && $self->{conn}->id || 'noid')." $reason\n";
     $self->close;
     return 0;
 }
@@ -88,6 +132,11 @@ sub mask {
     $fd->mask($rw,$v);
 }
 
+sub fd {
+    my ($self,$dir) = @_;
+    return $self->{fds}[$dir];
+}
+
 # send some data via fd dir
 sub forward {
     my ($self,$from,$to,$data) = @_;
@@ -97,6 +146,56 @@ sub forward {
     $fo->write($data,$from);
 }
 
+# ssl interception, e.g. upgrade both client and server to SSL sockets,
+# where I can read/write unencrypted data
+sub sslify {
+    my ($self,$from,$to,$hostname,$callback) = @_;
+    my $conn = $self->{conn} or return;
+    my $mitm = $conn->{mitm} or return; # no MITM needed
+
+    # destroy the current connection object and create a new obne
+    $conn = $self->{conn} = $conn->clone;
+    $conn->{intunnel} = 1;
+    
+    my $sfo = $self->{fds}[$from] or return
+	$self->fatal("cannot startssl $from - no such fo");
+
+    # stop handling all data
+    $self->mask($to,r=>0);
+    $self->mask($from,r=>0);
+    weaken( my $wself = $self );
+
+    my %sslargs = (
+	SSL_verifycn_name => $hostname,
+	SSL_verifycn_schema => 'http',
+	SSL_hostname => $hostname, # SNI
+	$conn->{capath} ? (
+	    SSL_verify_mode => SSL_VERIFY_PEER,
+	    ( -d $conn->{capath} ? 'SSL_ca_path' : 'SSL_ca_file' ), 
+	    $conn->{capath}
+	):( 
+	    SSL_verify_mode => SSL_VERIFY_NONE 
+	)
+    );
+    $sfo->startssl( %sslargs, sub {
+	my $sfo = shift;
+	my ($cert,$key) = $mitm->clone_cert($sfo->{fd}->peer_certificate);
+	my $cfo = $wself->{fds}[$to] or return
+	    $wself->fatal("cannot startssl $to - no such fo");
+	$cfo->startssl(
+	    SSL_server => 1,
+	    SSL_cert => $cert,
+	    SSL_key  => $key,
+	    sub {
+		# allow data again
+		$self->mask($to,r=>1);
+		$self->mask($from,r=>1);
+		$callback->() if $callback;
+	    }
+	);
+    });
+}
+
 # closes relay
 sub close:method {
     my $self = shift;
@@ -178,6 +277,7 @@ use Carp 'croak';
 use Scalar::Util 'weaken';
 use App::HTTP_Proxy_IMP::Debug;
 use AnyEvent::Socket qw(tcp_connect format_address);
+use IO::Socket::SSL;
 
 use fields (
     'dir',        # direction 0,1
@@ -215,7 +315,7 @@ sub xdebug {
 	unshift @_,$conn,$msg;
 	goto &$xdebug;
     } else {
-	debug(@_);
+	goto &debug;
     }
 }
 
@@ -275,33 +375,32 @@ sub shutdown:method {
 
 
 sub mask {
-    my ($self,$rw,$mask) = @_;
-    #debug("$self->{dir} $self->{fd} fn=".fileno($self->{fd})." $rw=>$mask");
+    my ($self,$rw,$val) = @_;
+    #debug("$self->{dir} $self->{fd} fn=".fileno($self->{fd})." $rw=>$val");
     if ( $rw eq 'r' ) {
-	if ( ! $mask ) {
+	if ( ! $val ) {
 	    # disable read
 	    undef $self->{rwatch};
-	} elsif ( ! $self->{rwatch} ) {
-	    # enable read if not enabled
+	} else {
 	    $self->{status} & 0b100 and return 0; # read shutdown already
 	    $self->{rsub} ||= sub { _read($self) }; 
 	    $self->{rwatch} = AnyEvent->io(
 		fh => $self->{fd},
 		poll => 'r',
-		cb => $self->{rsub}
+		cb => ref($val) ? $val : $self->{rsub}
 	    );
 	}
     } elsif ( $rw eq 'w' ) {
-	if ( ! $mask ) {
+	if ( ! $val ) {
 	    # disable write
 	    undef $self->{wwatch};
-	} elsif ( $self->{wbuf} ne '' and ! $self->{wwatch} ) {
+	} else {
 	    $self->{status} & 0b010 and return 0; # write shutdown already
 	    $self->{wsub} ||= sub { _writebuf($self) }; 
 	    $self->{wwatch} = AnyEvent->io(
 		fh => $self->{fd},
 		poll => 'w',
-		cb => $self->{wsub}
+		cb => ref($val) ? $val : $self->{wsub}
 	    );
 	}
     } else {
@@ -317,7 +416,7 @@ sub write:method {
     if ( $self->{wbuf} eq '' ) {
 	# no buffered data, set as buffer and try to write immediately
 	$self->{wbuf} = $data;
-	$n = _writebuf($self,$from);
+	$n = _writebuf($self,$from) // return; # fatal?
     } else {
 	# only append to buffer, will be written on write ready
 	$self->{wbuf} .= $data;
@@ -337,7 +436,7 @@ sub _writebuf {
     my $self = shift;
     #debug("write $self fn=".fileno($self->{fd}));
     my $n = syswrite($self->{fd},$self->{wbuf});
-    #debug("write done: ". (defined $n ? $n : $!));
+    #debug("write(%s,%d) -> %s", $self->{dir},length($self->{wbuf}), (defined $n ? $n : $!));
     if ( ! defined $n ) {
         $self->{relay}->fatal("write($self->{dir}) failed: $!")
 	    unless $!{EINTR} or $!{EAGAIN};
@@ -362,6 +461,10 @@ sub _writebuf {
 	if ( my $src = $self->{wsrc} ) {
 	    $self->{relay}->mask($_, r=>1) for (keys %$src);
 	}
+    } else {
+	# need to write more later
+	#debug("need to write more");
+	mask($self,w=>1);
     }
     return $n;
 }
@@ -437,6 +540,7 @@ sub connect:method {
 		    $callback->();
 		} else {
 		    App::HTTP_Proxy_IMP::Relay::DNS::uncache($host,$addr);
+		    $self->{relay} or return; # relay already closed
 		    $self->{relay}->fatal("connect to $host.$port failed: $!");
 		}
 	    });
@@ -448,6 +552,46 @@ sub connect:method {
     return -1;
 }
 
+sub startssl {
+    my $self = shift;
+    $self->{rbuf} eq '' or return 
+	$self->{relay}->fatal("read buf $self->{dir} not empty before starting SSL: '$self->{rbuf}'");
+    $self->{wbuf} eq '' or return 
+	$self->{relay}->fatal("write buf $self->{dir} not empty before starting SSL: '$self->{wbuf}'");
+
+    my $callback = @_%2 ? pop(@_):undef;
+    my %sslargs = @_;
+    IO::Socket::SSL->start_SSL( $self->{fd},
+	%sslargs,
+	SSL_startHandshake => 0,
+    ) or die "failed to upgrade socket to SSL";
+    my $sub = $sslargs{SSL_server} 
+	? \&IO::Socket::SSL::accept_SSL
+	: \&IO::Socket::SSL::connect_SSL;
+    _ssl($self,$sub,$callback,\%sslargs);
+}
+
+sub _ssl {
+    my ($self,$sub,$cb,$sslargs) = @_;
+    if ( $sub->($self->{fd}) ) {
+	$self->xdebug("ssl handshake success");
+	$cb->($self) if $cb;
+    } elsif ( $!{EAGAIN} ) {
+	# retry
+	my $dir = 
+	    $SSL_ERROR == SSL_WANT_READ ? 'r' :
+	    $SSL_ERROR == SSL_WANT_WRITE ? 'w' :
+	    return $self->{relay}->fatal( "unhandled $SSL_ERROR on EAGAIN" );
+	$self->mask( $dir => sub { _ssl($self,$sub,$cb,$sslargs) });
+    } elsif ( $sslargs->{SSL_server} ) {
+	return $self->{relay}->fatal( "error on accept_SSL: $SSL_ERROR|$!" );
+    } else {
+	return $self->{relay}->fatal( 
+	    "error on connect_SSL to $sslargs->{SSL_verifycn_name}: $SSL_ERROR|$!" );
+    }
+}
+
+
 ############################################################################
 # DNS cache
 ############################################################################
